/**
* @file calcCluster.cc
* @brief Wrapper program for overlapping community detection methods
* 
* Calculates the results of running a given overlapping community detection algorithm on a given network.
* 
* Input network should be undirected (no duplicate edges), and in an edge list format. Each line of the file should represent an edge with the following pattern:
*	 Vertex1 Vertex2 Weight
* 
* If multiple networks are to be run in the same function call, they should be suffixed with (positive) numbers in increasing order, the range of which is specified by the parameters
* 
* @author James Thompson
* @date Aug 7 2012 
* @version 1.2
*/

#include <boost/lexical_cast.hpp>
#include <libgen.h>

#include <stdarg.h>

#include <boost/timer.hpp>

#include "graph.h"
#include "cluster.h"
#include "density.h"

#include "CIS.h"

#include "nation.h"
/**
* @fn int main(int argc, char* argv[])
* @brief Main entry point
* 
* @param -alg ! Overlapping Community Detection Algorithm
* @arg \b CIS ::IterativeDensity "Connected Iterative Scan"
* @arg \b 3CL ::CliqueCluster "3-Clique Percolation"
* @arg \b CNG ::CONGA "Cluster Overlap Newman Girvan Algorithm"
* @arg \b COP ::COPRA "Community Overlap Propagation Algorithm"
* @arg \b COM ::LCOMM "Link Communities"
* @arg \b RRW ::RRW "Repeated Random Walks"
* @arg \b SLP ::SLPA "Speaker-Listener Label Propagation Algorithm"
* @arg \b SDE ::SSDE "Sampled Spectral Distance Embedding (Not Working)"
* @arg \b OSL ::OSLOM "Order Statistics Local Optimization Method"
*
* @param -args ! Arguments for specified community detection algorithm
* @arg \b CIS [lambda] [density type: c = complex, n = negative]
* @arg \b 3CL 
* @arg \b CNG [Locality of search (1+)] [number of clusters]
* @arg \b COP [Threshold Parameter]
* @arg \b COM 
* @arg \b RRW [restart Prob] [max size] [min size] [overlap threshold] [early cutoff]
* @arg \b SLP [Threshold] [Number of Runs] [iterations]
* @arg \b SDE ---------------------------------
* @arg \b OSL 
* 
* @param -i ! Network input file
* 
* @param -o Community results output file
* 
* @param -rLo Lowest suffix for range of network files
* 
* @param -rHi Highest suffix for range of network files
* 
* @param -tr Translate output back to original vertex names instead of outputting IDs
* 
* @param -d Delimiter(s) used between vertex names and weights in the network file \b default [space] [comma] and [tab]
* 
* @param -a Factor to multiply positive edges with
* 
* @param -b Factor to multiply negative edges with
* 
* @param -wt Weight Threshold - edges whose weights are below this number are discarded
* 
*/
int main(int argc, char *argv[])
{   
    //Variables to hold the values of all the parameters
    map <string, vector < string > > params = getParameters(argc, argv);
    map <string, vector < string > >::iterator it_pr;
    string ALG, IN, OUT, delimiters;
    double alpha = 1, beta = 1, weightThresh = 0.;
    bool translate = false, special = false;
    string ARGS;
    int rLo = -1, rHi = -1;
    
    //Help string
    string USAGE = boost::lexical_cast<string>(argv[0]) + " -alg=Algorithm -i=inputfile -o=output -args=AlgorithmArguments\n\tSee documentation for full list of flags\n";
    
    //Parse through the parameters
    try{
      if ((it_pr = params.find("alg")) != params.end()){
	ALG = (it_pr->second)[0];
	if (strcmp(ALG.c_str(), "CIS") == 0){
	  special = true;
	  USAGE += "\nArgs for CIS : [lambda value] [type of density (c - complex; n-negative)]";
	} else if (strcmp(ALG.c_str(), "RRW") == 0){
	  USAGE += "\nArgs for Random Walks: [restart Prob] [max size] [min size] [overlap threshold] [early cutoff]";
	} else if (strcmp(ALG.c_str(), "CNG") == 0){
	  USAGE += "\nArgs for CONGA : [Locality of search (1+)]";
	} else if (strcmp(ALG.c_str(), "SLP") == 0){
	  USAGE += "\nArgs for SLPA: [Threshold] [Number of Runs] [iterations]";
	} else if (strcmp(ALG.c_str(), "3CL") == 0){
	  USAGE += "\nArgs for 3CL: ";
	} else if (strcmp(ALG.c_str(), "OSL") == 0){
	  USAGE += "\nUsual args for OSL: -hr 0 -w";
	} else if (strcmp(ALG.c_str(), "COM") == 0){
	  USAGE += "\nArgs for COM: ";
	} else if (strcmp(ALG.c_str(), "COP") == 0){
	  USAGE += "\nArgs for COP: [Threshold Parameter]";
	} else if (strcmp(ALG.c_str(), "SDE") == 0){
	  USAGE += "\nArgs for SDE: ";
	}
      } else {
	cerr << USAGE << endl;
	exit(3);
      }
      
      if ((it_pr = params.find("i")) != params.end()){
	IN = (it_pr->second)[0];
      } else {
	cerr << USAGE << endl;
	exit(3);
      }
      
      if ((it_pr = params.find("o")) != params.end()){
	OUT = (it_pr->second)[0];
      } else {
	OUT = ALG + "_Results.dat";
      }
      
      if ((it_pr = params.find("rLo")) != params.end()){
	rLo = boost::lexical_cast<int>((it_pr->second)[0]);
      }
      
      if ((it_pr = params.find("rHi")) != params.end()){
	rHi = boost::lexical_cast<int>((it_pr->second)[0]);
      }
      
      if ((it_pr = params.find("args")) != params.end()){
	ARGS = it_pr->second[0];
      } else {
	ARGS = "";
      }
      
      if ((it_pr = params.find("tr")) != params.end()){
	translate = true;
      }
      
      if ((it_pr = params.find("d")) != params.end()){
	delimiters = it_pr->second[0];
      } else {
	delimiters = " \t,";
      }
      
      if ((it_pr = params.find("b")) != params.end()){
	beta = boost::lexical_cast<double>(it_pr->second[0]);
      }
      
      if ((it_pr = params.find("a")) != params.end()){
	alpha = boost::lexical_cast<double>(it_pr->second[0]);
      }
      
      if ((it_pr = params.find("wt")) != params.end()){
	weightThresh = boost::lexical_cast<double>(it_pr->second[0]);
      }
      
    } catch (boost::bad_lexical_cast &e) {
      cerr << USAGE << endl;
      exit(3);
    }

    
    //Go through each file to be clustered
    for (int reps = rLo; reps <= rHi; reps++){
      ClusterAlg<double>* ca;  /* Clustering algorithm of the type specified (new for each file since results are saved, could change to include a 'clear' method)*/
      ca = ClusterAlg<double>::Create(ALG.c_str(), ARGS, alpha, beta, weightThresh, translate);
      
      //Open the current Graph File to cluster
      ifstream fin;
      
      string currentFile = IN, currentOutFile = OUT;
      if (reps >= 0){
	currentFile += boost::lexical_cast<string>(reps);
	currentOutFile += boost::lexical_cast<string>(reps);
      }
      
      //Set output file / input file
      ca->setOutput(currentOutFile);
      openFile(&fin, currentFile.c_str());
      
      // Load Network
      if (special)
	((IterativeDensity<double>*)ca)->ReadGraph(&fin, delimiters);
      else
	ca->ReadGraph(&fin, delimiters);
      
      // Calculate clusters
      if (special)
	((IterativeDensity<double>*)ca)->calculate();
      else
	ca->calculate();
      
      // If output wasn't translated, output ID map for future use
      if (!translate){
	WriteMap<string, lui>(&(ca->idmap), (currentOutFile + "_ID").c_str());
      }
  
      //Clean up
      fin.close();
      //delete ca;
    }
}