/** 
 * @file cluster.h
 * @brief Overlapping Community Detection Algorithm Implementations
 * 
 * Describes a class for multiple community detection algorithms. Each separate class handles housekeeping / overhead for clustering algorithm,
 * 	including formatting the input, reformatting output, and formatting the command to run the algorithm.
 */

#ifndef CLUSTER_H_
#define CLUSTER_H_

#include <queue>
#include <vector>

#include <map>
#include <set>
#include <vector>
#include <fstream>
#include <iostream>

#include <stdarg.h>
#include "graph.h"
#include "density.h"
#include "seed.h"

#include "boost/tuple/tuple.hpp"
#include "boost/tuple/tuple_comparison.hpp"
#include "boost/tuple/tuple_io.hpp"

#include "ComponentsLoop.h"

namespace std {
template<class T >
class complex;}

using namespace std;

/** @class ClusterAlg
 * @brief Base class for all community detection algorithms
 * 
 * Holds general purpose functions for all algorithms, but main purpose is as a factory, to create the correct class at runtime.
 * 
 * @arg R The type of weight used in the network (usually int or double - using double is safer)
 */
template <typename R>
class ClusterAlg
{
  public:
    /** Will be the actual algorithm implementation. */
    virtual void calculate() = 0;
 
    static density<bidirected_t>* dens; /**< Community density calculation method to use*/
    
  /** Retrieve the density of a set of vertices in a graph using boost structures. 
   * @param graph The network the cluster is from
   * @param cluster Set of vertices in the cluster
   * @return Cluster density value
   */
  float get_Density(bidirected_t::graph* graph, const set < bidirected_t::vertex >* cluster)
  {      
      set < bidirected_t::vertex >::iterator it;
      it = cluster->begin();
      
      //Create a seed with the first vertex
      seed calc(dens, *graph, *it);
      
      //Add the remaining vertices to the seed.	
      for (++it; it != cluster->end(); ++it)
      {
	calc.add_vertex(*it);
      }
	
      //Calculate density
      return calc.dens();
  }
  
  /** Retrieve density of a set of vertices 
   * 
   * @param cluster IDs of vertices in the community
   * @return Cluster density value
   */
  float get_Density(set <lui>* cluster){
    double win = 0, wout = 0;
    
    set <lui>::iterator it_sl;
    typename map <lui, map <lui, R> >::iterator it_gr;
    typename map <lui, R>::iterator it_nei;
    
    //Calculate internal and external link weights
    sui k, t;
    for (it_sl = cluster->begin(), k = 0; it_sl != cluster->end() && k < cluster->size(); it_sl++, k++){
	if ((it_gr = network.find(*it_sl)) != network.end()){
	  for (it_nei = it_gr->second.begin(), t = 0; it_nei != it_gr->second.end() && t < it_gr->second.size(); it_nei++, t++){
	    if (cluster->find(it_nei->first) != cluster->end()){
	      win += it_nei->second;
	    } else {
	      wout += it_nei->second;
	    }
	  }
	}
    }
   
    //Density calculation
    return (*(this->dens))(win, wout, cluster->size());
  }
  
  /** 
   * Factory function
   * 
   * @param type Algorithm to create class for
   * @param argList Arguments to use for algorithm
   * @param alpha Factor for positive weights
   * @param beta Factor for negative weights
   * @param weightThresh Minimum edge weight
   * 
   * @return Cluster algorithm set up to implement 'type' algorithm
   */
  static ClusterAlg* Create (string type, string args, double alpha, double beta, double weightThresh, bool translate);
  
  /**
   * Read network from a file and save. Each vertex read in is given a numerical ID that it used throughout the detection process. (Easier
   * comparison between numbers than strings). This can be translated back later.
   * 
   * @param fin Stream to read network from
   * @param delimiters Delimiters between the vertex IDs and edge weights in the network file.
   * 
   * @return true
   */
  bool ReadGraph(ifstream* fin, string delimiters = " ,\t", double alpha = 1, double beta = 1) {
      vector <string> fields;
      map < string, lui >::iterator it_id;
      lui nxtID = 1;
      typename map < lui, map < lui, R > >::iterator it_gr, it_gr2;
      lui source, target;
     
      while (fline_tr(fin, &fields, delimiters.c_str())){
      if (fields.size() != 3) continue;
     
      //Check to see if vertex already has ID. If not, give the next ID to vertex.
      if ((it_id = idmap.find(fields[0]))!=idmap.end()){
	source = it_id->second;
      } else {
	idmap.insert(make_pair<string, lui>(fields[0], nxtID));
	source = nxtID++;
      } 
      
      //Same for second edge endpoint
      if ((it_id = idmap.find(fields[1]))!=idmap.end()){
	target = it_id->second;
      } else {
	idmap.insert(make_pair<string, lui>(fields[1], nxtID));
	target = nxtID++;
      }
      
      R weight = boost::lexical_cast<R>(fields[2]);
      if (weight > 0)
	weight = alpha * weight;
      else
	weight = beta * weight;
      
      //Add the edge to the network      
      if ((it_gr = network.find(source)) != network.end()){
	if ((it_gr2 = network.find(target)) != network.end()){
	  if (it_gr2->second.find(source) != it_gr2->second.end()){
	    //The edge is already in the network
	    cerr << "Duplicate edge : " << source << " - " << target << " :: " << fields[2] << endl;
	  } else {
	    it_gr->second.insert(make_pair<lui,R>(target, weight));
	  }
	} else {
	  it_gr->second.insert(make_pair<lui,R>(target, weight));
	}
      } else if ((it_gr2 = network.find(target)) != network.end()){
	it_gr2->second.insert(make_pair<lui,R>(source, weight));
      } else {
	map < lui, R > tmp;
	tmp.insert(make_pair<lui,R>(target, weight));
	network.insert(make_pair<lui, map <lui, R> >(source, tmp));
      }
    }
      
    for (it_id = this->idmap.begin(); it_id != this->idmap.end(); it_id++){
      idLookup.insert(make_pair<lui, string>(it_id->second, it_id->first));
    }
    return true;
  };
  
  /**
   * Print the network to a file
   * 
   * @param fout Stream to print file to
   * 
   * @return true
   */
  bool PrintGraph(ofstream* fout){
    typename map < lui, map < lui, R > >::iterator source;
    typename map < lui, R >::iterator target;
    
    for (source = network.begin(); source != network.end(); source++){
      for (target = source->second.begin(); target != source->second.end(); target++){
	(*fout) << source->first << " " << target->first << " " << target->second << endl;
      }
    }
    return true;
  }
  
  /**
   * Print detected communities to file. Each line in the output file is as follows:
   * 
   * CommunityDensity CommunitySize Vertex1ID Vertex2ID Vertex3ID ...
   * 
   * @param filename String representing filename to print communities to
   * @return true
   */
  bool PrintCommunities(string filename){
    ofstream fout(filename.c_str());
    set < lui >::iterator it_cl;
    
    for (int i = 0; i < communities.size(); i++){
      fout << this->get_Density(&communities[i]) << " " << communities[i].size();
      for (it_cl = communities[i].begin(); it_cl != communities[i].end(); it_cl++){
	fout << " " << *it_cl;
      }
      fout << endl;
    }
    
    return true;
  }
  
  /**
   * Print a single community to the out file
   * 
   * CommunityDensity CommunitySize Vertex1ID Vertex2ID Vertex3ID ...
   * 
   * @param comm Set of vertices to print out
   */
  void PrintCommunity(set <lui> * comm){
    set < lui >::iterator it_cl;
    
    *fileout << this->get_Density(&(*comm)) << " " << (*comm).size();
    for (it_cl = (*comm).begin(); it_cl != (*comm).end(); it_cl++){
      *fileout << " " << *it_cl;
    }
    *fileout << endl;
  }
  
  /**
   * Print detected communities to file. Each line in the output file is as follows:
   * 
   * CommunityDensity CommunitySize Vertex1 Vertex2 Vertex3 ...
   * 
   * @param filename String representing filename to print communities to
   * @return true
   */
  bool PrintTranslatedCommunities(string filename){
    set <lui>::iterator it_sl; 
    
    ofstream fout(filename.c_str());
    set < lui >::iterator it_cl;
    
    for (int i = 0; i < communities.size(); i++){
      fout << this->get_Density(&communities[i]) << " " << communities[i].size();
      for (it_cl = communities[i].begin(); it_cl != communities[i].end(); it_cl++){
	fout << " " << (idLookup.find(*it_cl))->second;
      }
    }
    
    fout << endl;
    
    return true;
  }
  
  /**
   * Print a single community to the out file (with translated names)
   * 
   * CommunityDensity CommunitySize Vertex1_OriginalID Vertex2_OriginalID Vertex3_OriginalID ...
   * 
   * @param comm Set of vertices to print out
   */
  void PrintTranslatedCommunity(set <lui> * comm){
    set < lui >::iterator it_cl;
    
    *fileout << this->get_Density(&(*comm)) << " " << (*comm).size();
    for (it_cl = (*comm).begin(); it_cl != (*comm).end(); it_cl++){
      *fileout << " " << (idLookup.find(*it_cl))->second;
    }
    *fileout << endl;
  }
  
  /**
   * As long as the original vertex IDs were also numerical (integers), this function will retrieve the original vertex IDs without printing
   * out the communities. Useful for comparing results to benchmark 'ground truth's.
   */
  void TranslateCommunities(){
    
    set < lui >::iterator it_cl;
    
    vector < set <lui> > copy = communities;
    communities.clear();
    
    for (int i = 0; i < copy.size(); i++){
      set <lui> clust;
      for (it_cl = copy[i].begin(); it_cl != copy[i].end(); it_cl++){
	clust.insert(boost::lexical_cast<lui>((idLookup.find(*it_cl))->second));
      }
      communities.push_back(clust);
    }    
  }
  
  void setOutput(string filename){
      fileout = new ofstream(filename.c_str());
  }
  
  map <lui, map <lui, R> > network; /**< Structure to hold network*/
  vector < set < lui > > communities; /**< Structure to hold resulting communities */
  map <string, lui> idmap; /**< Holds map to original vertex names */
  map <lui, string> idLookup; /**< For faster translating*/
  
  string args; /**< Clustering Algorithm Arguments */
  
  bool translate;/**< @see ClusterAlg::Create */
  ostream* fileout; /**< Place to print output*/  
};

#endif
