/**
 * @file seed.cc
 * @brief Controls a set of vertices being considered for community detection in CIS
 */

#include "seed.h"
#include <cassert>

seed :: seed(const density<bidirected_t>* dens, const bidirected_t::graph& g):
    g_(&g), dens_(dens), win_(0), wout_(0), vertices_()
{
}

	/*
	 * The vertex s is the first node to consider as part of the 'cluster'
	 */
seed ::seed(const density<bidirected_t>* dens, const bidirected_t::graph& g,
        bidirected_t::vertex s):
    g_(&g), dens_(dens), win_(0), wout_(0), vertices_()
{
    bidirected_t::edge_name_t p_plus  = 0;
    bidirected_t::edge_name_t p_minus = 0;
    bidirected_t::edge_name_t p_not   = 0;

    find_p(s, &p_plus, &p_minus, &p_not);

    vertices_.insert(s);

    // Our new internal weight is equal to our old internal weight, plus all
    // edges originating from v to vertices_ (p_plus) plus all edges originating
    // from vertices_ to v (p_minus).
    win_ = win_ + p_plus + p_minus;
    assert(win_ == 0); // No internal edges with just one vertex
    // Our new external weight is equal to the old external weight, minus all
    // edges that have been converted to internal edges (p_minus), plus all
    // edges that originate from v to outside the set (p_not).
    wout_ = wout_ - (p_plus + p_minus) + p_not;
    assert(wout_ == p_not); // Since it's the first node, there should be no internal edges (p_minus = 0)
}


seed :: seed(const seed& other):
    g_(other.g_), dens_(other.dens_), win_(other.win_), wout_(other.wout_),
    vertices_(other.vertices_)
{
}


seed&
seed :: operator =(const seed& rhs)
{
    if (this == &rhs)
    {
        return *this;
    }
    g_        = rhs.g_;
    dens_     = rhs.dens_;
    win_      = rhs.win_;
    wout_     = rhs.wout_;
    vertices_ = rhs.vertices_;
    return *this;
}


/*	Calculates the density of vertices_ both with and without v added. Returns true if the density
 * with v added is greater than that of the original.
 */
bool
seed :: can_add_vertex(const bidirected_t::vertex& v) const
{
    if (vertices_.find(v) != vertices_.end())
    {
        // Vertex already added
        return false;
    }

    lui k = vertices_.size();
    bidirected_t::edge_name_t p_plus  = 0;
    bidirected_t::edge_name_t p_minus = 0;
    bidirected_t::edge_name_t p_not   = 0;

    find_p(v, &p_plus, &p_minus, &p_not);

    // Our new internal weight is equal to our old internal weight, plus all
    // edges originating from v to vertices_ (p_plus) plus all edges originating
    // from vertices_ to v (p_minus).
    bidirected_t::edge_name_t new_win = win_ + p_plus + p_minus;
    // Our new external weight is equal to the old external weight, minus all
    // edges that have been converted to internal edges (p_minus), plus all
    // edges that originate from v to outside the set (p_not).
    bidirected_t::edge_name_t new_wout = wout_ - (p_minus + p_plus) + p_not;

    double old_dens = (*dens_)(win_, wout_, k);
    double new_dens = (*dens_)(new_win, new_wout, k + 1);

    return old_dens <= new_dens; // [Maybe <= ?] No, equalities handled by threshold and post-work
}

//Just like can_add_vertex, except the 'new_dens' is now with v removed.
bool
seed :: can_rm_vertex(const bidirected_t::vertex& v) const
{
    if (vertices_.find(v) == vertices_.end())
    {
        // Vertex already removed
        return false;
    }

    lui k = vertices_.size();
    bidirected_t::edge_name_t p_plus  = 0;
    bidirected_t::edge_name_t p_minus = 0;
    bidirected_t::edge_name_t p_not   = 0;

    find_p(v, &p_plus, &p_minus, &p_not);

    // Our new internal edge weight is equal to the old internal edge weight
    // minus all edges originating from v to vertices (p_plus) minus all edges
    // originating from vertices to v (p_minus)
    bidirected_t::edge_name_t new_win = win_ - (p_plus + p_minus);
    // Our new external edge weight is equal to the old external edge weight
    // plus all edges originating from vertices to v (p_minus) minus all edges
    // originating from v to the outside.
    bidirected_t::edge_name_t new_wout = wout_ + (p_plus + p_minus) - p_not;

    double old_dens = (*dens_)(win_, wout_, k);
    double new_dens = (*dens_)(new_win, new_wout, k - 1);
    
    bool Fractured = false;
    /*rm_vertex(v);
    
    std::vector<seed>* pieces = fracture_seed(*this);
    if (pieces->size() > 1)
      Fractured = true;
    
    add_vertex(v);*/

    return ((old_dens < new_dens) && !Fractured);
}


void
seed :: add_vertex(const bidirected_t::vertex& v)
{
    bidirected_t::edge_name_t p_plus  = 0;
    bidirected_t::edge_name_t p_minus = 0;
    bidirected_t::edge_name_t p_not   = 0;

    find_p(v, &p_plus, &p_minus, &p_not);

    // Our new internal weight is equal to our old internal weight, plus all
    // edges originating from v to vertices_ (p_plus) plus all edges originating
    // from vertices_ to v (p_minus).
    win_ = win_ + p_plus + p_minus;
    // Our new external weight is equal to the old external weight, minus all
    // edges that have been converted to internal edges (p_minus), plus all
    // edges that originate from v to outside the set (p_not).
    wout_ = wout_ - (p_minus + p_plus) + p_not;

    /*typename bidirected_t::vertex_name_t u;
    u = boost::get(boost::vertex_name, *g_, v);
    cerr << "Added " << u <<  endl;*/
    vertices_.insert(v);
}


void
seed :: rm_vertex(const bidirected_t::vertex& v)
{
    bidirected_t::edge_name_t p_plus  = 0;
    bidirected_t::edge_name_t p_minus = 0;
    bidirected_t::edge_name_t p_not   = 0;

    find_p(v, &p_plus, &p_minus, &p_not);

    // Our new internal edge weight is equal to the old internal edge weight
    // minus all edges originating from v to vertices (p_plus) minus all edges
    // originating from vertices to v (p_minus)
    win_ = win_ - (p_plus + p_minus);
    // Our new external edge weight is equal to the old external edge weight
    // plus all edges originating from vertices to v (p_minus) minus all edges
    // originating from v to the outside.
    wout_ = wout_ + (p_plus + p_minus) - p_not;

    /*typename bidirected_t::vertex_name_t u;
    u = boost::get(boost::vertex_name, *g_, v);
    cerr << "Removed " << u <<  endl;*/
    vertices_.insert(v);
    
    vertices_.erase(v);
}


/*
 * Sets the values of p_plus, p_minus and p_not to (respectively) the value of edges starting from vertex
 *	v and staying within the current set (vertices_), edges that start inside the current set and are connected to vertex v, 
 *	and edges that start at vertex v and are connected to vertices outside the current set.
 */
void
seed ::find_p(const bidirected_t::vertex& v, bidirected_t::edge_name_t* p_plus,
        bidirected_t::edge_name_t* p_minus,
        bidirected_t::edge_name_t* p_not) const
{
    boost::graph_traits<bidirected_t::graph>::in_edge_iterator is, ie;
    boost::graph_traits<bidirected_t::graph>::out_edge_iterator os, oe;

    *p_plus  = 0;
    *p_minus = 0;
    *p_not   = 0;
    
    vertex_set::iterator it;
    
    boost::tie(os, oe) = boost::out_edges(v, *g_);
	/*Loops through all the edges leaving vertex v using iterators*/
    for (; os != oe; ++os)
    {
		/*	
		 *	If the vertex on the other end of the edge is not in the given list of vertices, then the
		 * edge's value is added to p_not, if the edge is found, the value is added to p_plus.
		 * If find fails, it returns the same as end()
		 */
        if (vertices_.find(boost::target(*os, *g_)) == vertices_.end())
        {
	  *p_not += boost::get(boost::edge_name, *g_, *os);
        }
        else
        {
            *p_plus += boost::get(boost::edge_name, *g_, *os);
        }
    }
	/*
	 *	Same type of loops, but p_minus counts the edges that start within the vertex group and connect to
	 * other vertices. In other words, it's the edges that go from being half internal half external to an edge
	 * that is all internal with the addition of vertex v.
	 */
    for (boost::tie(is, ie) = boost::in_edges(v, *g_); is != ie; ++is)
    {
        if (vertices_.find(boost::source(*is, *g_)) != vertices_.end())
        {
            *p_minus += boost::get(boost::edge_name, *g_, *is);
        }
        else
	{
	  *p_not += boost::get(boost::edge_name, *g_, *is);
	}
    }
}

void seed::print()
{
  vertex_set::iterator it;
  
  std::cerr << "|| Current Seed ||" << std::endl;
  
  for (it = vertices_.begin(); it != vertices_.end(); it++)
  {
    std::cerr << "Author: " << boost::get(boost::vertex_name, *g_, *it) << std::endl;
  }
  
  std::cerr << dens() << std::endl;
}

/*
 * The vector returned consists of all subgraphs in seed S where no edge crosses between two subgraphs.
 */
std::vector<seed>* fracture_seed(const seed& S)
{
    std::vector<seed>* pieces = new std::vector<seed>();
    seed::vertex_set seen;
	
	/*Goes through each vertex in the seed and if it hasn't already been 'seen', 
	 * Calls the subroutine on it.
	 */
    for (seed::vertex_set::const_iterator i = S.vertices().begin();
            i != S.vertices().end(); ++i)
    {
        if (seen.find(*i) == seen.end())
        {
            seen.insert(*i);
            pieces->push_back(seed(S.dens_func(), S.graph(), *i));
            fracture_seed(S, seen, pieces->back(), *i);
        }
    }
    return pieces;
}

/*Fills the seed P with all the vertices connected to v that are in seed S but have not been flagged
 * in 'seen' yet
 */
void fracture_seed(const seed& S, seed::vertex_set& seen,
        seed& P, bidirected_t::vertex v)
{
    boost::graph_traits<bidirected_t::graph>::in_edge_iterator is, ie;
    boost::graph_traits<bidirected_t::graph>::out_edge_iterator os, oe;

	/*Goes through the edges of the current vertex*/
    for (boost::tie(os, oe) = boost::out_edges(v, S.graph()); os != oe; ++os)
    {
        bidirected_t::vertex u = boost::target(*os, S.graph());
	/*If the target is part of the seed, but hasn't been flagged, flag it, add it to the 
	 * active fragment, and then recurse.
	 */
        if (S.vertices().find(u) != S.vertices().end() &&
                seen.find(u) == seen.end())
        {
            seen.insert(u);
            P.add_vertex(u);
            fracture_seed(S, seen, P, u);
        }
    }
    for (boost::tie(is, ie) = boost::in_edges(v, S.graph()); is != ie; ++is)
    {
        bidirected_t::vertex u = boost::source(*is, S.graph());
        if (S.vertices().find(u) != S.vertices().end() &&
                seen.find(u) == seen.end())
        {
            seen.insert(u);
            P.add_vertex(u);
            fracture_seed(S, seen, P, u);
        }
    }
}